"""
Task Manager for handling asynchronous tasks.

This module provides the `TaskManager` class, which manages a collection of `asyncio.Task`
objects. It ensures proper tracking and cleanup of tasks, handles exceptions,
and supports context management for ease of use.
"""

from __future__ import annotations

import asyncio
import dataclasses

from . import logconfig


@dataclasses.dataclass
class TaskManager:
    """
    Manages a collection of asyncio Tasks, ensuring they are tracked and cleaned up appropriately.

    The `TaskManager` class allows you to add tasks to a managed set, automatically
    handles exceptions from tasks, and removes tasks from the set when they are done.
    It also provides asynchronous context management to facilitate clean startup and
    shutdown of tasks within an `async with` block.

    Attributes:
        tasks (set[asyncio.Task]): A set of asyncio Task objects being managed.
    """

    tasks: set[asyncio.Task] = dataclasses.field(
        default_factory=set,
        init=False,
    )

    def log_unhandled_exception(self, task: asyncio.Task) -> None:
        """
        Log any unhandled exceptions from a completed task.

        This method is a callback for tasks that have finished execution.
        If the task resulted in an exception, it logs the error using the application's logger.

        Args:
            task (asyncio.Task): The task that has completed.
        """
        if exception := task.exception():
            logconfig.logger.error(
                "Unhandled exception in task: %s",
                task,
                exc_info=exception,
            )

    def add(self, task: asyncio.Task) -> None:
        """
        Add an asyncio Task to the manager and set up callbacks.

        Registers the task with the manager's task set and adds callbacks to handle
        unhandled exceptions and to remove the task from the set when it is done.

        Args:
            task (asyncio.Task): The asyncio Task to be managed.
        """
        self.tasks.add(task)
        task.add_done_callback(self.log_unhandled_exception)
        task.add_done_callback(self.tasks.remove)

    async def gather_tasks(self, return_exceptions: bool = True) -> list[BaseException | None]:
        """
        Wait for all managed tasks to complete and gather their results.

        This method awaits all tasks currently in the manager's task set,
        collecting their results or exceptions. It ensures that all tasks
        have finished before proceeding.

        Returns:
            list[BaseException | None]: A list containing exceptions raised by tasks,
                or None for tasks that completed successfully.
        """
        return await asyncio.gather(
            *self.tasks,
            return_exceptions=return_exceptions,
        )

    async def __aenter__(self) -> TaskManager:
        """
        Enter the asynchronous context manager.

        Returns:
            TaskManager: The instance of the TaskManager itself.
        """
        return self

    async def __aexit__(self, *_: object) -> None:
        """
        Exit the asynchronous context manager, ensuring all tasks are gathered.

        Upon exiting the context, this method waits for all managed tasks to complete
        by calling `gather_tasks()`. This ensures that no tasks are left running when
        the context is exited.
        """
        await self.gather_tasks()



from __future__ import annotations

import asyncio
import random
from dataclasses import dataclass
from datetime import timedelta
from itertools import count

import pytest

from pgqueuer.db import Driver
from pgqueuer.models import Job
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


@dataclass
class Tally:
    count: int


async def consumer(qm: QueueManager, tally: Tally) -> None:
    @qm.entrypoint("asyncfetch", requests_per_second=100)
    async def asyncfetch(job: Job) -> None:
        tally.count += 1

    @qm.entrypoint("syncfetch", requests_per_second=10)
    def syncfetch(job: Job) -> None:
        tally.count += 1

    await qm.run(dequeue_timeout=timedelta(seconds=0))


async def enqueue(
    queries: Queries,
    size: int,
) -> None:
    assert size > 0
    cnt = count()
    entrypoints = ["syncfetch", "asyncfetch"] * size
    await queries.enqueue(
        random.sample(entrypoints, k=size),
        [f"{next(cnt)}".encode() for _ in range(size)],
        [0] * size,
    )


@pytest.mark.parametrize("concurrency", (1, 2, 3, 4))
async def test_rps(
    concurrency: int,
    apgdriver: Driver,
    n_tasks: int = 1_000,
    wait: int = 5,
) -> None:
    tally = Tally(count=0)

    await enqueue(Queries(apgdriver), size=n_tasks)

    qms = [QueueManager(apgdriver) for _ in range(concurrency)]

    async def dequeue() -> None:
        consumers = [consumer(q, tally) for q in qms]
        await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    await asyncio.gather(timer(), dequeue())
    assert 100 <= tally.count / wait <= 140



import asyncio
import functools
import uuid
from datetime import datetime, timedelta, timezone
from multiprocessing import Process, Queue as MPQueue

import pytest

from pgqueuer.db import Driver
from pgqueuer.executors import EntrypointExecutor, JobExecutor, is_async_callable
from pgqueuer.models import Job
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


class MultiprocessingExecutor(JobExecutor):
    def __init__(self) -> None:
        self.requests_per_second = 5
        self.retry_timer = timedelta(seconds=10)
        self.serialized_dispatch = False
        self.concurrency_limit = 2
        self.queue: MPQueue[object] = MPQueue()

    async def execute(self, job: Job) -> None:
        process = Process(target=self.process_function, args=(job,))
        process.start()
        process.join()

    def process_function(self, job: Job) -> None:
        # Simulate processing and put result in queue
        if job.payload:
            self.queue.put(job.payload)


@pytest.mark.asyncio
async def test_entrypoint_executor_sync() -> None:
    result = []

    def sync_function(job: Job) -> None:
        if job.payload:
            result.append(job.payload)

    executor = EntrypointExecutor(func=sync_function)
    now = datetime.now(timezone.utc)
    job = Job(
        id=1,
        priority=1,
        created=now,
        heartbeat=now,
        status="queued",
        entrypoint="test",
        payload=b"test_payload",
        queue_manager_id=uuid.uuid4(),
    )

    await executor.execute(job)

    assert result == [b"test_payload"]


@pytest.mark.asyncio
async def test_entrypoint_executor_async() -> None:
    result = []

    async def async_function(job: Job) -> None:
        if job.payload:
            result.append(job.payload)

    executor = EntrypointExecutor(func=async_function)
    now = datetime.now(timezone.utc)
    job = Job(
        id=1,
        priority=1,
        created=now,
        heartbeat=now,
        status="queued",
        entrypoint="test",
        payload=b"test_payload",
        queue_manager_id=uuid.uuid4(),
    )

    await executor.execute(job)

    assert result == [b"test_payload"]


@pytest.mark.asyncio
async def test_custom_threading_executor() -> None:
    class ThreadingExecutor(JobExecutor):
        def __init__(self) -> None:
            self.requests_per_second = 10
            self.retry_timer = timedelta(seconds=5)
            self.serialized_dispatch = False
            self.concurrency_limit = 5
            self.result = list[bytes]()

        async def execute(self, job: Job) -> None:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self.threaded_function, job)

        def threaded_function(self, job: Job) -> None:
            if job.payload:
                self.result.append(job.payload)

    executor = ThreadingExecutor()
    now = datetime.now(timezone.utc)
    job = Job(
        id=1,
        priority=1,
        created=now,
        heartbeat=now,
        status="queued",
        entrypoint="test",
        payload=b"thread_payload",
        queue_manager_id=uuid.uuid4(),
    )

    await executor.execute(job)

    assert executor.result == [b"thread_payload"]


@pytest.mark.asyncio
async def test_custom_multiprocessing_executor() -> None:
    executor = MultiprocessingExecutor()
    now = datetime.now(timezone.utc)
    job = Job(
        id=1,
        priority=1,
        created=now,
        heartbeat=now,
        status="queued",
        entrypoint="test",
        payload=b"process_payload",
        queue_manager_id=uuid.uuid4(),
    )

    await executor.execute(job)
    result = executor.queue.get()

    assert result == b"process_payload"


@pytest.mark.asyncio
async def test_queue_manager_with_custom_executor(apgdriver: Driver) -> None:
    qm = QueueManager(connection=apgdriver)
    results = []

    class CustomExecutor(JobExecutor):
        async def execute(self, job: Job) -> None:
            if job.payload:
                results.append(job.payload)

    @qm.entrypoint(name="custom_entrypoint", executor=CustomExecutor)
    def entrypoint_function(job: Job) -> None:
        pass  # Not used since executor handles execution

    queries = Queries(apgdriver)
    await queries.enqueue(entrypoint="custom_entrypoint", payload=b"test_data")

    async def run_queue_manager() -> None:
        await qm.run(dequeue_timeout=timedelta(seconds=2), batch_size=1)

    task = asyncio.create_task(run_queue_manager())
    await asyncio.sleep(1)  # Allow some time for the job to be processed
    qm.shutdown.set()  # Signal the queue manager to stop
    await task

    assert results == [b"test_data"]


async def async_function(job: Job) -> None:
    await asyncio.sleep(0)


def sync_function(job: Job) -> None:
    pass


def test_is_async_callable_with_async_function() -> None:
    assert is_async_callable(async_function) is True


def test_is_async_callable_with_sync_function() -> None:
    assert is_async_callable(sync_function) is False


def test_is_async_callable_with_partial_async_function() -> None:
    partial_async = functools.partial(async_function)
    assert is_async_callable(partial_async) is True


def test_is_async_callable_with_partial_sync_function() -> None:
    partial_sync = functools.partial(sync_function)
    assert is_async_callable(partial_sync) is False


def test_is_async_callable_with_class_method() -> None:
    class MyClass:
        async def async_method(self, job: Job) -> None:
            await asyncio.sleep(0)

        def sync_method(self, job: Job) -> None:
            pass

    instance = MyClass()
    assert is_async_callable(instance.async_method) is True
    assert is_async_callable(instance.sync_method) is False



import asyncio
import time
import uuid
from datetime import timedelta

import pytest

from pgqueuer import db
from pgqueuer.models import Job
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


async def wait_until_empty_queue(
    q: Queries,
    qms: list[QueueManager],
) -> None:
    while sum(x.count for x in await q.queue_size()) > 0:
        await asyncio.sleep(0.01)

    for qm in qms:
        qm.shutdown.set()


@pytest.mark.parametrize("N", (1, 2, 32))
async def test_job_queing(
    apgdriver: db.Driver,
    N: int,
) -> None:
    c = QueueManager(apgdriver)
    seen = list[int]()

    @c.entrypoint("fetch")
    async def fetch(context: Job) -> None:
        assert context.payload is not None
        seen.append(int(context.payload))

    await c.queries.enqueue(
        ["fetch"] * N,
        [f"{n}".encode() for n in range(N)],
        [0] * N,
    )

    await asyncio.gather(
        c.run(),
        wait_until_empty_queue(c.queries, [c]),
    )

    assert seen == list(range(N))


@pytest.mark.parametrize("N", (1, 2, 32))
@pytest.mark.parametrize("concurrency", (1, 2, 3, 4))
async def test_job_fetch(
    apgdriver: db.Driver,
    N: int,
    concurrency: int,
) -> None:
    q = Queries(apgdriver)
    qmpool = [QueueManager(apgdriver) for _ in range(concurrency)]
    seen = list[int]()

    for qm in qmpool:

        @qm.entrypoint("fetch")
        async def fetch(context: Job) -> None:
            assert context.payload is not None
            seen.append(int(context.payload))

    await q.enqueue(
        ["fetch"] * N,
        [f"{n}".encode() for n in range(N)],
        [0] * N,
    )

    await asyncio.gather(
        asyncio.gather(*[qm.run() for qm in qmpool]),
        wait_until_empty_queue(q, qmpool),
    )

    assert sorted(seen) == list(range(N))


@pytest.mark.parametrize("N", (1, 2, 32))
@pytest.mark.parametrize("concurrency", (1, 2, 3, 4))
async def test_sync_entrypoint(
    apgdriver: db.Driver,
    N: int,
    concurrency: int,
) -> None:
    q = Queries(apgdriver)
    qmpool = [QueueManager(apgdriver) for _ in range(concurrency)]
    seen = list[int]()

    for qm in qmpool:

        @qm.entrypoint("fetch")
        def fetch(context: Job) -> None:
            time.sleep(1)  # Sim. heavy CPU/IO.
            assert context.payload is not None
            seen.append(int(context.payload))

    await q.enqueue(
        ["fetch"] * N,
        [f"{n}".encode() for n in range(N)],
        [0] * N,
    )

    await asyncio.gather(
        asyncio.gather(*[qm.run() for qm in qmpool]),
        wait_until_empty_queue(q, qmpool),
    )
    assert sorted(seen) == list(range(N))


async def test_pick_local_entrypoints(
    apgdriver: db.Driver,
    N: int = 100,
) -> None:
    q = Queries(apgdriver)
    qm = QueueManager(apgdriver)
    pikced_by = list[str]()

    @qm.entrypoint("to_be_picked")
    async def to_be_picked(job: Job) -> None:
        pikced_by.append(job.entrypoint)

    await q.enqueue(["to_be_picked"] * N, [None] * N, [0] * N)
    await q.enqueue(["not_picked"] * N, [None] * N, [0] * N)

    async def waiter() -> None:
        while sum(x.count for x in await q.queue_size() if x.entrypoint == "to_be_picked"):
            await asyncio.sleep(0.01)
        qm.shutdown.set()

    await asyncio.gather(
        qm.run(dequeue_timeout=timedelta(seconds=0.01)),
        waiter(),
    )

    assert pikced_by == ["to_be_picked"] * N
    assert sum(s.count for s in await q.queue_size() if s.entrypoint == "to_be_picked") == 0
    assert sum(s.count for s in await q.queue_size() if s.entrypoint == "not_picked") == N


async def test_pick_set_queue_manager_id(
    apgdriver: db.Driver,
    N: int = 100,
) -> None:
    q = Queries(apgdriver)
    qm = QueueManager(apgdriver)
    qmids = set[uuid.UUID]()

    @qm.entrypoint("fetch")
    async def fetch(job: Job) -> None:
        assert job.queue_manager_id is not None
        qmids.add(job.queue_manager_id)

    await q.enqueue(["fetch"] * N, [None] * N, [0] * N)

    async def waiter() -> None:
        while sum(x.count for x in await q.queue_size()):
            await asyncio.sleep(0.01)
        qm.shutdown.set()

    await asyncio.gather(
        qm.run(dequeue_timeout=timedelta(seconds=0.01)),
        waiter(),
    )

    assert len(qmids) == 1



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from datetime import timedelta
from itertools import count

import async_timeout
import pytest

from pgqueuer.db import Driver
from pgqueuer.models import Job, JobId
from pgqueuer.qm import QueueManager
from pgqueuer.queries import DBSettings, Queries


async def _inspect_queue_jobs(jids: list[JobId], driver: Driver) -> list[Job]:
    sql = f"""SELECT * FROM {DBSettings().queue_table} WHERE id = ANY($1::integer[])"""
    return [Job.model_validate(x) for x in await driver.fetch(sql, jids)]


@dataclass
class Tally:
    active: int = 0
    max_active: int = 0
    lock: asyncio.Lock = field(default_factory=asyncio.Lock)

    async def __aenter__(self) -> None:
        async with self.lock:
            self.active += 1
            self.max_active = max(self.max_active, self.active)

    async def __aexit__(self, *_: object) -> None:
        async with self.lock:
            self.active -= 1


async def consumer(qm: QueueManager, tally: Tally, limit: int) -> None:
    @qm.entrypoint("fetch", concurrency_limit=limit)
    async def fetch(job: Job) -> None:
        async with tally:
            await asyncio.sleep(0.001)

    await qm.run(dequeue_timeout=timedelta(seconds=0))


async def enqueue(
    queries: Queries,
    size: int,
    cnt: count = count(),
) -> None:
    assert size > 0
    await queries.enqueue(
        ["fetch"] * size,
        [f"{next(cnt)}".encode() for _ in range(size)],
        [0] * size,
    )


@pytest.mark.parametrize("n_consumers", (1, 4))
@pytest.mark.parametrize("max_concurrency", (1, 5, 10))
async def test_max_concurrency(
    n_consumers: int,
    max_concurrency: int,
    apgdriver: Driver,
    n_tasks: int = 500,
    wait: int = 1,
) -> None:
    await enqueue(Queries(apgdriver), size=n_tasks)

    tallys = [Tally() for _ in range(n_consumers)]
    qms = [QueueManager(apgdriver) for _ in range(n_consumers)]

    async def dequeue() -> None:
        consumers = [consumer(q, tally, max_concurrency) for q, tally in zip(qms, tallys)]
        await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    await asyncio.gather(timer(), dequeue())

    for tally in tallys:
        assert max_concurrency - 1 <= tally.max_active <= max_concurrency


@pytest.mark.parametrize("concurrency_limit", (1, 5, 10))
@pytest.mark.parametrize("batch_size", (1, 5, 10))
@pytest.mark.parametrize("dequeue_timeout", (1, 5, 10))
async def test_concurrency_entrypoint_isolation(
    apgdriver: Driver,
    concurrency_limit: int,
    batch_size: int,
    dequeue_timeout: float,
) -> None:
    event = asyncio.Event()
    N = concurrency_limit * 1_000
    await Queries(apgdriver).enqueue(
        ["fetch_1", "fetch_2"] * N,
        [None, None] * N,
        [0, 0] * N,
    )

    qm = QueueManager(apgdriver)

    @qm.entrypoint("fetch_1", concurrency_limit=concurrency_limit)
    async def fetch_1(job: Job) -> None:
        await event.wait()

    @qm.entrypoint("fetch_2", concurrency_limit=concurrency_limit)
    async def fetch_2(job: Job) -> None:
        await event.wait()

    async def timer() -> None:
        async with async_timeout.timeout(10):
            while len(event._waiters) < 2 * concurrency_limit:
                await asyncio.sleep(0.001)
            len_waiter = len(event._waiters)
            qm.shutdown.set()
            event.set()
            assert len_waiter == 2 * concurrency_limit

    await asyncio.gather(
        timer(),
        qm.run(
            batch_size=batch_size,
            dequeue_timeout=timedelta(seconds=dequeue_timeout),
        ),
    )



"""
Module for initializing and managing PostgreSQL event listeners.

This module defines the `PGNoticeEventListener` class, which extends `asyncio.Queue`
to store events received from a PostgreSQL NOTIFY channel. It also provides the
`initialize_notice_event_listener` function to set up the listener and handle
different types of events, such as table changes, request rates, and job cancellations.
"""

from __future__ import annotations

import asyncio

from . import db, logconfig, models


class PGNoticeEventListener(asyncio.Queue[models.TableChangedEvent]):
    """
    Asynchronous queue for PostgreSQL events from a specified channel.

    The `PGNoticeEventListener` class is a specialized `asyncio.Queue` that stores
    `TableChangedEvent` instances received from a PostgreSQL NOTIFY channel.
    It allows consumers to asynchronously retrieve events as they are received.
    """


async def initialize_notice_event_listener(
    connection: db.Driver,
    channel: models.PGChannel,
    statistics: dict[str, models.EntrypointStatistics],
    canceled: dict[models.JobId, models.Context],
) -> PGNoticeEventListener:
    """
    Initialize a listener on a PostgreSQL channel to handle various events.

    Sets up a listener on the specified PostgreSQL NOTIFY channel using the provided
    database connection. When notifications are received, it parses the payloads
    and handles different types of events, such as table changes, requests per second,
    and job cancellations.

    Args:
        connection (db.Driver): The database driver instance used to add the listener.
        channel (models.PGChannel): The name of the PostgreSQL channel to listen on.
        statistics (dict[str, models.EntrypointStatistics]): A dictionary to store
            request rate statistics for entrypoints.
        canceled (dict[models.JobId, models.Context]): A mapping of job IDs to their
            cancellation contexts.

    Returns:
        PGNoticeEventListener: An instance of `PGNoticeEventListener` that queues
            received `TableChangedEvent` instances.
    """

    def parse_and_queue(
        payload: str | bytes | bytearray,
        notice_event_queue: PGNoticeEventListener,
        statistics: dict[str, models.EntrypointStatistics],
    ) -> None:
        """
        Parse a notification payload and handle the event accordingly.

        Parses the JSON payload received from the PostgreSQL NOTIFY channel and processes
        it based on the event type. It supports handling table change events, requests
        per second events, and cancellation events.

        Args:
            payload (str | bytes | bytearray): The raw payload from the notification.
            notice_event_queue (PGNoticeEventListener): The event queue to put table
                change events into.
            statistics (dict[str, models.EntrypointStatistics]): A dictionary to update
                with requests per second statistics.
        """
        try:
            parsed = models.AnyEvent.model_validate_json(payload)
        except Exception as e:
            logconfig.logger.critical("Failed to parse payload: `%s`", payload, exc_info=e)
            return

        if parsed.root.type == "table_changed_event":
            notice_event_queue.put_nowait(parsed.root)
        elif parsed.root.type == "requests_per_second_event":
            statistics[parsed.root.entrypoint].samples.append(
                (parsed.root.count, parsed.root.sent_at)
            )
        elif parsed.root.type == "cancellation_event":
            for jid in parsed.root.ids:
                if ctx := canceled.get(jid):
                    ctx.cancellation.cancel()
        else:
            raise NotImplementedError(parsed, payload)

    notice_event_listener = PGNoticeEventListener()

    await connection.add_listener(
        channel,
        lambda x: parse_and_queue(x, notice_event_listener, statistics),
    )

    return notice_event_listener



import asyncio
import asyncio.selector_events
from collections import Counter
from datetime import datetime, timedelta

import async_timeout
import pytest

from pgqueuer import db, queries
from pgqueuer.models import Job, JobId
from pgqueuer.qm import QueueManager


async def _inspect_queue_jobs(jids: list[JobId], driver: db.Driver) -> list[Job]:
    sql = f"""SELECT * FROM {queries.DBSettings().queue_table} WHERE id = ANY($1::integer[])"""
    return [Job.model_validate(x) for x in await driver.fetch(sql, jids)]


@pytest.mark.parametrize("N", (2, 4, 8, 16))
async def test_retry(
    apgdriver: db.Driver,
    N: int,
    retry_timer: timedelta = timedelta(seconds=0.100),
) -> None:
    e = asyncio.Event()
    c = QueueManager(apgdriver)
    jobids = Counter[JobId]()

    @c.entrypoint("fetch", retry_timer=retry_timer)
    async def fetch(context: Job) -> None:
        jobids[context.id] += 1
        await e.wait()

    await c.queries.enqueue(["fetch"] * N, [None] * N, [0] * N)

    async def stop_after() -> None:
        async with async_timeout.timeout(retry_timer.total_seconds() * 10):
            while True:
                if len(jobids) == N and all(v > 1 for v in jobids.values()):
                    break
                await asyncio.sleep(0)

        e.set()
        c.shutdown.set()

    await asyncio.gather(
        c.run(dequeue_timeout=timedelta(seconds=0)),
        stop_after(),
    )
    assert len(jobids) == N
    assert all(v > 1 for v in jobids.values())


async def test_no_retry_on_zero_timer(
    apgdriver: db.Driver,
    retry_timer: timedelta = timedelta(seconds=0),
) -> None:
    e = asyncio.Event()
    c = QueueManager(apgdriver)
    seen = set[datetime]()

    @c.entrypoint("fetch", retry_timer=retry_timer)
    async def fetch(context: Job) -> None:
        seen.add(context.heartbeat)
        await e.wait()

    await c.queries.enqueue(["fetch"], [None], [0])

    async def until_retry_updated() -> None:
        await asyncio.sleep(1)
        c.shutdown.set()
        e.set()

    await asyncio.gather(
        c.run(dequeue_timeout=timedelta(seconds=0)),
        until_retry_updated(),
    )

    assert len(seen) == 1


async def test_heartbeat_updates(
    apgdriver: db.Driver,
    retry_timer: timedelta = timedelta(seconds=0.100),
) -> None:
    e = asyncio.Event()
    c = QueueManager(apgdriver)
    seen = []

    @c.entrypoint("fetch", retry_timer=retry_timer)
    async def fetch(context: Job) -> None:
        seen.append(context.heartbeat)
        await e.wait()

    jids = await c.queries.enqueue(["fetch"], [None], [0])
    (before,) = await _inspect_queue_jobs(jids, apgdriver)
    after: None | Job = None

    async def until_retry_updated() -> None:
        nonlocal after
        while len(seen) < 2:
            await asyncio.sleep(0)
        (after,) = await _inspect_queue_jobs(jids, apgdriver)
        c.shutdown.set()
        e.set()

    async with async_timeout.timeout(retry_timer.total_seconds() * 10):
        await asyncio.gather(
            c.run(dequeue_timeout=timedelta(seconds=0)),
            until_retry_updated(),
        )

    assert after is not None
    assert after.heartbeat > before.heartbeat
    assert len(set(seen)) > 1


@pytest.mark.parametrize("N", (2, 4))
async def test_concurrent_retry_jobs(
    apgdriver: db.Driver,
    N: int,
    retry_timer: timedelta = timedelta(seconds=0.100),
) -> None:
    e1 = asyncio.Event()
    e2 = asyncio.Event()
    c = QueueManager(apgdriver)
    seen_job1 = Counter[JobId]()
    seen_job2 = Counter[JobId]()

    @c.entrypoint("fetch1", retry_timer=retry_timer)
    async def fetch1(context: Job) -> None:
        seen_job1[context.id] += 1
        await e1.wait()

    @c.entrypoint("fetch2", retry_timer=retry_timer)
    async def fetch2(context: Job) -> None:
        seen_job2[context.id] += 1
        await e2.wait()

    await c.queries.enqueue(["fetch1", "fetch2"] * N, [None, None] * N, [0, 0] * N)

    async def until_retry_updated() -> None:
        while True:
            if (
                len(seen_job1) == N
                and all(v > 1 for v in seen_job1.values())
                and len(seen_job2) == N
                and all(v > 1 for v in seen_job2.values())
            ):
                break
            await asyncio.sleep(0)

        c.shutdown.set()
        e1.set()
        e2.set()

    async with async_timeout.timeout(retry_timer.total_seconds() * 2 * N):
        await asyncio.gather(
            c.run(dequeue_timeout=timedelta(seconds=0)),
            until_retry_updated(),
        )

    assert len(seen_job1) == N
    assert all(v > 1 for v in seen_job1.values())

    assert len(seen_job2) == N
    assert all(v > 1 for v in seen_job2.values())


@pytest.mark.parametrize("N", (2, 4))
async def test_varying_retry_timers(
    apgdriver: db.Driver,
    N: int,
    retry_timer_short: timedelta = timedelta(seconds=0.200),
    retry_timer_long: timedelta = timedelta(seconds=0.400),
) -> None:
    e_short = asyncio.Event()
    e_long = asyncio.Event()
    c = QueueManager(apgdriver)
    seen_short = Counter[JobId]()
    seen_long = Counter[JobId]()

    @c.entrypoint("fetch_short", retry_timer=retry_timer_short)
    async def fetch_short(context: Job) -> None:
        seen_short[context.id] += 1
        await e_short.wait()

    @c.entrypoint("fetch_long", retry_timer=retry_timer_long)
    async def fetch_long(context: Job) -> None:
        seen_long[context.id] += 1
        await e_long.wait()

    await c.queries.enqueue(["fetch_short", "fetch_long"], [None, None], [0, 0])

    async def until_retry_updated() -> None:
        while True:
            if (
                seen_short
                and all(v > 1 for v in seen_short.values())
                and seen_long
                and all(v > 1 for v in seen_long.values())
            ):
                break
            await asyncio.sleep(0)

        c.shutdown.set()
        e_short.set()
        e_long.set()

    async with async_timeout.timeout(2):
        await asyncio.gather(
            c.run(dequeue_timeout=timedelta(seconds=0)),
            until_retry_updated(),
        )

    assert seen_short and all(v > 1 for v in seen_short.values())
    assert seen_long and all(v > 1 for v in seen_long.values())


@pytest.mark.parametrize("N", (2, 4))
async def test_retry_with_cancellation(
    apgdriver: db.Driver,
    N: int,
    retry_timer: timedelta = timedelta(seconds=0.100),
) -> None:
    e = asyncio.Event()
    c = QueueManager(apgdriver)
    seen = set[datetime]()
    call_count = 0

    @c.entrypoint("fetch", retry_timer=retry_timer)
    async def fetch(context: Job) -> None:
        nonlocal call_count
        seen.add(context.heartbeat)
        call_count += 1
        if call_count < N:
            raise asyncio.CancelledError("Simulated cancellation")
        await e.wait()

    await c.queries.enqueue(["fetch"], [None], [0])

    async def until_retry_updated() -> None:
        while len(seen) < N:
            await asyncio.sleep(0)
        c.shutdown.set()
        e.set()

    async with async_timeout.timeout(retry_timer.total_seconds() * 2 * N):
        await asyncio.gather(
            c.run(dequeue_timeout=timedelta(seconds=0)),
            until_retry_updated(),
        )

    assert call_count == N
    assert (
        pytest.approx(
            (max(seen) - min(seen)).total_seconds(),
            abs=retry_timer.total_seconds(),
        )
        == (N * retry_timer).total_seconds()
    )



"""
Queue Manager module for handling job queues and dispatching jobs.

This module defines the `QueueManager` class, which manages job queues, dispatches
jobs to registered entrypoints, and handles database connections and events.
It also provides utility functions and types for job execution and rate limiting.
"""

from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import sys
import uuid
import warnings
from collections import Counter, deque
from datetime import timedelta
from math import isfinite
from typing import AsyncGenerator, Callable

import anyio

from . import buffers, db, executors, heartbeat, helpers, listeners, logconfig, models, queries, tm


class Unset: ...


@dataclasses.dataclass
class QueueManager:
    """
    Manages job queues and dispatches jobs to registered entrypoints.

    The `QueueManager` handles the lifecycle of jobs, including dequeueing,
    dispatching to entrypoint functions, handling rate limiting, concurrency
    limits, and managing cancellations.

    Attributes:
        connection (db.Driver): The database driver used for database operations.
        channel (models.PGChannel): The PostgreSQL channel for notifications.
        shutdown (asyncio.Event): Event to signal when the QueueManager is shutting down.
        queries (queries.Queries): Instance for executing database queries.
        entrypoint_registry (dict[str, JobExecutor]): Registered job executors.
        entrypoint_statistics (dict[str, models.EntrypointStatistics]): Statistics for entrypoints.
        queue_manager_id (uuid.UUID): Unique identifier for each QueueManager instance.
        job_context (dict[models.JobId, models.Context]): Contexts for jobs,
            including cancellation scopes.
    """

    connection: db.Driver
    channel: models.PGChannel = dataclasses.field(
        default=models.PGChannel(queries.DBSettings().channel),
    )

    shutdown: asyncio.Event = dataclasses.field(
        init=False,
        default_factory=asyncio.Event,
    )
    queries: queries.Queries = dataclasses.field(init=False)

    # Per entrypoint
    entrypoint_registry: dict[str, executors.JobExecutor] = dataclasses.field(
        init=False,
        default_factory=dict,
    )
    entrypoint_statistics: dict[str, models.EntrypointStatistics] = dataclasses.field(
        init=False,
        default_factory=dict,
    )
    queue_manager_id: uuid.UUID = dataclasses.field(
        init=False,
        default_factory=uuid.uuid4,
    )

    # Per job.
    job_context: dict[models.JobId, models.Context] = dataclasses.field(
        init=False,
        default_factory=dict,
    )

    @property
    def alive(self) -> asyncio.Event:
        # For backwards compatibility
        warnings.warn(
            "The `alive` property is deprecated and will be removed in a future release. "
            "Please use `shutdown` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.shutdown

    def __post_init__(self) -> None:
        """
        Initialize the QueueManager after dataclass fields have been set.

        Sets up the `queries` instance using the provided database connection.
        """
        self.queries = queries.Queries(self.connection)

    def get_context(self, job_id: models.JobId) -> models.Context:
        """
        Retrieve the context associated with a specific job ID.

        Args:
            job_id (models.JobId): The unique identifier of the job.

        Returns:
            models.Context: The context object associated with the job.
        """
        return self.job_context[job_id]

    def register_executor(
        self,
        name: str,
        executor: executors.JobExecutor,
    ) -> None:
        """
        Register a job executor with a specific name.

        Args:
            name (str): The name of the entrypoint as referenced in the job queue.
            executor (JobExecutor): The job executor instance.

        Raises:
            RuntimeError: If the entrypoint name is already registered.
        """
        if name in self.entrypoint_registry:
            raise RuntimeError(f"{name} already in registry, name must be unique.")

        self.entrypoint_registry[name] = executor
        self.entrypoint_statistics[name] = models.EntrypointStatistics(
            samples=deque(maxlen=1_000),
            concurrency_limiter=asyncio.Semaphore(executor.concurrency_limit or sys.maxsize),
        )

    def entrypoint(
        self,
        name: str,
        *,
        requests_per_second: float = float("inf"),
        concurrency_limit: int = 0,
        retry_timer: timedelta = timedelta(seconds=0),
        serialized_dispatch: bool = False,
        executor: type[executors.JobExecutor] = executors.EntrypointExecutor,
    ) -> Callable[[executors.EntrypointTypeVar], executors.EntrypointTypeVar]:
        """
        Decorator to register an entrypoint for job processing.

        Users can specify a custom executor or use the default EntrypointExecutor.

        Args:
            name (str): The name of the entrypoint as referenced in the job queue.
            requests_per_second (float): Max number of jobs per second to process for
                this entrypoint.
            concurrency_limit (int): Max number of concurrent jobs allowed for this entrypoint.
            retry_timer (timedelta): Duration to wait before retrying 'picked' jobs.
            serialized_dispatch (bool): Whether to serialize dispatching of jobs.
            executor (JobExecutor): Custom executor instance to use.

        Returns:
            Callable[[T], T]: A decorator that registers the function as an entrypoint.

        Raises:
            RuntimeError: If the entrypoint name is already registered.
            ValueError: If `requests_per_second` or `concurrency_limit` are negative.
        """

        if name in self.entrypoint_registry:
            raise RuntimeError(f"{name} already in registry, name must be unique.")

        # Check requests_per_second type / value range.
        if not isinstance(requests_per_second, (float, int)):
            raise ValueError("Rate must be float | int.")

        if requests_per_second < 0:
            raise ValueError("Rate must be greater or eq. to zero.")

        # Check concurrency_limit type / value range.
        if not isinstance(concurrency_limit, int):
            raise ValueError("Concurrency limit must be int.")

        if concurrency_limit < 0:
            raise ValueError("Concurrency limit must be greater or eq. to zero.")

        # Check serialized_dispatch type.
        if not isinstance(serialized_dispatch, bool):
            raise ValueError("Serialized dispatch must be boolean")

        def register(func: executors.EntrypointTypeVar) -> executors.EntrypointTypeVar:
            self.register_executor(
                name,
                executor(
                    func=func,
                    requests_per_second=requests_per_second,
                    retry_timer=retry_timer,
                    serialized_dispatch=serialized_dispatch,
                    concurrency_limit=concurrency_limit,
                ),
            )
            return func

        return register

    def observed_requests_per_second(
        self,
        entrypoint: str,
        epsilon: timedelta = timedelta(milliseconds=0.01),
    ) -> float:
        """
        Calculate the observed requests per second for an entrypoint.

        Args:
            entrypoint (str): The entrypoint to calculate the rate for.
            epsilon (timedelta): Small time delta to prevent division by zero.

        Returns:
            float: The observed requests per second.
        """
        samples = self.entrypoint_statistics[entrypoint].samples
        if not samples:
            return 0.0
        timespan = helpers.perf_counter_dt() - min(t for _, t in samples) + epsilon
        requests = sum(c for c, _ in samples)
        return requests / timespan.total_seconds()

    def entrypoints_below_capacity_limits(self) -> set[str]:
        """
        Determine which entrypoints are below their configured capacity limits.

        Returns:
            set[str]: A set of entrypoint names that are below capacity.
        """
        return {
            entrypoint
            for entrypoint, executor in self.entrypoint_registry.items()
            if self.observed_requests_per_second(entrypoint) < executor.requests_per_second
            and not self.entrypoint_statistics[entrypoint].concurrency_limiter.locked()
        }

    async def fetch_jobs(
        self,
        batch_size: int,
        task_manager: tm.TaskManager,
    ) -> AsyncGenerator[models.Job, None]:
        """
        Fetch jobs from the queue that are ready for processing, yielding them one at a time.

        This method retrieves a batch of jobs that match the current capacity constraints
        of each entrypoint. It ensures that jobs are fetched only when the QueueManager
        is operational and that rate limits and concurrency controls are respected.
        """

        while not self.shutdown.is_set():
            entrypoints = {
                x: (
                    self.entrypoint_registry[x].retry_timer,
                    self.entrypoint_registry[x].serialized_dispatch,
                    self.entrypoint_registry[x].concurrency_limit,
                )
                for x in self.entrypoints_below_capacity_limits()
            }

            if not (
                jobs := await self.queries.dequeue(
                    batch_size=batch_size,
                    entrypoints=entrypoints,
                    queue_manager_id=self.queue_manager_id,
                )
            ):
                break

            for job in jobs:
                yield job

            for entrypoint, count in Counter(job.entrypoint for job in jobs).items():
                # skip if rate is inf.
                target_rps = self.entrypoint_registry[entrypoint].requests_per_second
                if isfinite(target_rps):
                    task_manager.add(
                        asyncio.create_task(
                            self.queries.notify_debounce_event(
                                entrypoint,
                                count,
                            )
                        )
                    )

    async def run(
        self,
        dequeue_timeout: timedelta = timedelta(seconds=30),
        batch_size: int = 10,
        retry_timer: timedelta | None | Unset = Unset(),
    ) -> None:
        """
        Run the main loop to process jobs from the queue.

        Continuously listens for events and dispatches jobs, managing connections
        and tasks, logging timeouts, and resetting connections upon termination.

        Args:
            dequeue_timeout (timedelta): Timeout duration for waiting to dequeue jobs.
            batch_size (int): Number of jobs to retrieve in each batch.
            retry_timer (timedelta | None): Duration to wait before retrying 'picked' jobs.

        Raises:
            RuntimeError: If required database columns or types are missing.
        """

        if not isinstance(retry_timer, Unset):
            warnings.warn(
                "retry_timer is deprecated, use retry_timer with the entrypoint decorator.",
                DeprecationWarning,
            )

        if not (await self.queries.has_updated_column()):
            raise RuntimeError(
                f"The {self.queries.qb.settings.queue_table} table is missing the "
                "updated column, please run 'python3 -m pgqueuer upgrade'"
            )

        if not (
            await self.queries.has_user_type(
                "canceled", self.queries.qb.settings.statistics_table_status_type
            )
        ):
            raise RuntimeError(
                f"The {self.queries.qb.settings.statistics_table_status_type} is missing the "
                "'canceled' type, please run 'python3 -m pgqueuer upgrade'"
            )

        if not (await self.queries.has_heartbeat_column()):
            raise RuntimeError(
                f"The {self.queries.qb.settings.queue_table} table is missing the "
                "heartbeat column, please run 'python3 -m pgqueuer upgrade'"
            )

        if not (await self.queries.has_queue_manager_id_column()):
            raise RuntimeError(
                f"The {self.queries.qb.settings.queue_table} table is missing the "
                "queue_manager_id column, please run 'python3 -m pgqueuer upgrade'"
            )

        job_status_log_buffer_timeout = timedelta(seconds=0.01)
        heartbeat_buffer_timeout = helpers.retry_timer_buffer_timeout(
            [x.retry_timer for x in self.entrypoint_registry.values()]
        )
        async with (
            buffers.JobStatusLogBuffer(
                max_size=batch_size,
                timeout=job_status_log_buffer_timeout,
                callback=self.queries.log_jobs,
            ) as jbuff,
            buffers.HeartbeatBuffer(
                max_size=sys.maxsize,
                timeout=heartbeat_buffer_timeout / 2,
                callback=self.queries.notify_activity,
            ) as hbuff,
            tm.TaskManager() as task_manager,
            self.connection,
        ):
            notice_event_listener = await listeners.initialize_notice_event_listener(
                self.connection,
                self.channel,
                self.entrypoint_statistics,
                self.job_context,
            )

            shutdown_task = asyncio.create_task(self.shutdown.wait())

            while not self.shutdown.is_set():
                async for job in self.fetch_jobs(batch_size, task_manager):
                    self.job_context[job.id] = models.Context(
                        cancellation=anyio.CancelScope(),
                    )

                    task_manager.add(
                        asyncio.create_task(
                            self._dispatch(
                                job,
                                jbuff,
                                hbuff,
                            )
                        )
                    )

                    with contextlib.suppress(asyncio.QueueEmpty):
                        notice_event_listener.get_nowait()

                event_task = helpers.wait_for_notice_event(
                    notice_event_listener,
                    dequeue_timeout,
                )
                await asyncio.wait(
                    (shutdown_task, event_task),
                    return_when=asyncio.FIRST_COMPLETED,
                )

        if not event_task.done():
            event_task.cancel()

        if not shutdown_task.done():
            shutdown_task.cancel()

    async def _dispatch(
        self,
        job: models.Job,
        jbuff: buffers.JobStatusLogBuffer,
        hbuff: buffers.HeartbeatBuffer,
    ) -> None:
        """
        Dispatch a job to its associated entrypoint executor.

        Handles asynchronous job execution, logs exceptions, updates job status,
        and manages job context including cancellation.

        Args:
            job (models.Job): The job to dispatch.
            jbuff (buffers.JobStatusLogBuffer): Buffer to log job completion status.
            hbuff (buffers.HeartbeatBuffer): Buffer to manage heartbeats.
        """

        logconfig.logger.debug(
            "Dispatching entrypoint/id: %s/%s",
            job.entrypoint,
            job.id,
        )

        executor = self.entrypoint_registry[job.entrypoint]

        async with (
            heartbeat.Heartbeat(
                job.id,
                executor.retry_timer,
                hbuff,
            ),
            self.entrypoint_statistics[job.entrypoint].concurrency_limiter,
        ):
            try:
                # Run the job unless it has already been cancelled. Check this here because jobs
                # can be cancelled between when they are dequeued and when they acquire the
                # concurrency limit semaphore.
                ctx = self.get_context(job.id)
                if not ctx.cancellation.cancel_called:
                    await executor.execute(job)
            except Exception:
                logconfig.logger.exception(
                    "Exception while processing entrypoint/job-id: %s/%s",
                    job.entrypoint,
                    job.id,
                )
                await jbuff.add((job, "exception"))
            else:
                logconfig.logger.debug(
                    "Dispatching entrypoint/id: %s/%s - successful",
                    job.entrypoint,
                    job.id,
                )
                canceled = ctx.cancellation.cancel_called
                await jbuff.add((job, "canceled" if canceled else "successful"))
            finally:
                self.job_context.pop(job.id, None)



import json
from contextlib import asynccontextmanager
from typing import AsyncGenerator

import asyncpg
from fastapi import Depends, FastAPI, Request, Response

from pgqueuer.db import AsyncpgDriver
from pgqueuer.queries import Queries


def get_pgq_queries(request: Request) -> Queries:
    """Retrieve Queries instance from FastAPI app context."""
    return request.app.extra["pgq_queries"]


def create_app() -> FastAPI:
    """
    Create and configure a FastAPI app with a lifespan context manager
    to handle database connection.
    """

    @asynccontextmanager
    async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
        """Manage async database connection throughout the app's lifespan."""
        connection = await asyncpg.connect()
        app.extra["pgq_queries"] = Queries(AsyncpgDriver(connection))
        try:
            yield
        finally:
            await connection.close()

    app = FastAPI(lifespan=lifespan)

    @app.get("/reset_password_email")
    async def reset_password_email(
        user_name: str,
        pgq_queries: Queries = Depends(get_pgq_queries),
    ) -> Response:
        """Enqueue a job to reset a user's password, identified by user_name."""
        await pgq_queries.enqueue(
            "reset_email_by_user_name",
            payload=json.dumps({"user_name": user_name}).encode(),
        )
        return Response(status_code=201)

    return app


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(create_app())



# executors.py

from __future__ import annotations

import asyncio
import functools
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import (
    Awaitable,
    Callable,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
    overload,
)

import anyio
import anyio.to_thread

from . import models

AsyncEntrypoint: TypeAlias = Callable[[models.Job], Awaitable[None]]
SyncEntrypoint: TypeAlias = Callable[[models.Job], None]
Entrypoint: TypeAlias = AsyncEntrypoint | SyncEntrypoint
EntrypointTypeVar = TypeVar("EntrypointTypeVar", bound=Entrypoint)


@overload
def is_async_callable(obj: AsyncEntrypoint) -> TypeGuard[AsyncEntrypoint]: ...


@overload
def is_async_callable(obj: SyncEntrypoint) -> TypeGuard[SyncEntrypoint]: ...


def is_async_callable(obj: object) -> bool:
    """
    Determines whether an object is an asynchronous callable.
    """
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
        callable(obj) and asyncio.iscoroutinefunction(obj.__call__)
    )


class JobExecutor(ABC):
    """
    Abstract base class for job executors.

    Users can subclass this to create custom job executors.
    """

    def __init__(
        self,
        func: Entrypoint,
        requests_per_second: float,
        retry_timer: timedelta,
        serialized_dispatch: bool,
        concurrency_limit: int,
    ):
        super().__init__()
        self.func = func
        self.requests_per_second = requests_per_second
        self.retry_timer = retry_timer
        self.serialized_dispatch = serialized_dispatch
        self.concurrency_limit = concurrency_limit

    @abstractmethod
    async def execute(self, job: models.Job) -> None:
        """
        Execute the given job.

        Args:
            job (models.Job): The job to execute.
        """


class EntrypointExecutor(JobExecutor):
    """
    Job executor that wraps an entrypoint function.

    Executes the provided function when processing a job.
    """

    def __init__(
        self,
        func: Entrypoint,
        requests_per_second: float = float("inf"),
        retry_timer: timedelta = timedelta(seconds=0),
        serialized_dispatch: bool = False,
        concurrency_limit: int = 0,
    ) -> None:
        self.func = func
        self.requests_per_second = requests_per_second
        self.retry_timer = retry_timer
        self.serialized_dispatch = serialized_dispatch
        self.concurrency_limit = concurrency_limit
        self.is_async = is_async_callable(func)

    async def execute(self, job: models.Job) -> None:
        """
        Execute the job using the wrapped function.

        Args:
            job (models.Job): The job to execute.
        """
        if self.is_async:
            await cast(AsyncEntrypoint, self.func)(job)
        else:
            await anyio.to_thread.run_sync(cast(SyncEntrypoint, self.func), job)



from __future__ import annotations

import argparse
import asyncio
import os
from datetime import timedelta
from typing import Literal

from tabulate import tabulate, tabulate_formats

from . import db, listeners, models, queries, supervisor


async def display_stats(
    log_stats: list[models.LogStatistics],
    tablefmt: str,
) -> None:
    print(
        tabulate(
            [
                (
                    stat.created.astimezone(),
                    stat.count,
                    stat.entrypoint,
                    stat.time_in_queue,
                    stat.status,
                    stat.priority,
                )
                for stat in log_stats
            ],
            headers=[
                "Created",
                "Count",
                "Entrypoint",
                "Time in Queue (HH:MM:SS)",
                "Status",
                "Priority",
            ],
            tablefmt=tablefmt,
        )
    )


async def display_pg_channel(
    connection: db.Driver,
    channel: models.PGChannel,
) -> None:
    listener = await listeners.initialize_notice_event_listener(
        connection,
        channel,
        {},
        {},
    )
    while True:
        print(repr(await listener.get()))


async def fetch_and_display(
    queries: queries.Queries,
    interval: timedelta | None,
    tail: int,
    tablefmt: str,
) -> None:
    clear_and_home = "\033[2J\033[H"
    while True:
        print(clear_and_home, end="")
        await display_stats(await queries.log_statistics(tail), tablefmt)
        if interval is None:
            return
        await asyncio.sleep(interval.total_seconds())


def cliparser() -> argparse.Namespace:
    common_arguments = argparse.ArgumentParser(
        add_help=False,
        prog="pgqueuer",
    )

    common_arguments.add_argument(
        "--prefix",
        default="",
        help=(
            "Specify a prefix for all pgqueuer tables, functions, etc. "
            "This helps in avoiding conflicts if running multiple instances. "
            "(If set, additional config is required.)"
        ),
    )

    common_arguments.add_argument(
        "--driver",
        default="apg",
        help="Specify the PostgreSQL driver to be used: asyncpg (apg) or psycopg (psy).",
        choices=["apg", "psy"],
    )

    common_arguments.add_argument(
        "--pg-dsn",
        help=(
            "Provide the connection string in the libpq URI format, including host, port, user, "
            "database, password, passfile, and SSL options. Must be properly quoted; "
            "IPv6 addresses must be in brackets. Example: postgres://user:pass@host:port/database. "
            "Defaults to the PGDSN environment variable if set."
        ),
        default=os.environ.get("PGDSN"),
    )

    common_arguments.add_argument(
        "--pg-host",
        help=(
            "Specify the database host address, which can be an IP or domain name. "
            "Defaults to the PGHOST environment variable if set."
        ),
        default=os.environ.get("PGHOST"),
    )

    common_arguments.add_argument(
        "--pg-port",
        help=(
            "Specify the port number for the server host. "
            "Defaults to the PGPORT environment variable or 5432 if not set."
        ),
        default=os.environ.get("PGPORT", "5432"),
    )

    common_arguments.add_argument(
        "--pg-user",
        help=(
            "Specify the database role for authentication. "
            "Defaults to the PGUSER environment variable if set."
        ),
        default=os.environ.get("PGUSER"),
    )

    common_arguments.add_argument(
        "--pg-database",
        help=(
            "Specify the name of the database to connect to. "
            "Defaults to the PGDATABASE environment variable if set."
        ),
        default=os.environ.get("PGDATABASE"),
    )

    common_arguments.add_argument(
        "--pg-password",
        help=(
            "Specify the password for authentication. "
            "Defaults to the PGPASSWORD environment variable if set."
        ),
        default=os.environ.get("PGPASSWORD"),
    )

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        prog="pgqueuer",
        description=(
            "PGQueuer command-line interface for managing the "
            "PostgreSQL-backed job queue system."
        ),
    )

    subparsers = parser.add_subparsers(
        dest="command", required=True, help="Available commands for managing pgqueuer."
    )

    subparsers.add_parser(
        "install",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[common_arguments],
        help="Install the necessary database schema for PGQueuer.",
    ).add_argument(
        "--dry-run",
        action="store_true",
        help=(
            "Print the SQL statements that would be executed without actually "
            "applying any changes to the database."
        ),
    )

    subparsers.add_parser(
        "uninstall",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[common_arguments],
        help="Remove the PGQueuer schema from the database.",
    ).add_argument(
        "--dry-run",
        action="store_true",
        help=(
            "Print the SQL statements that would be executed without "
            "actually applying any changes to the database."
        ),
    )

    subparsers.add_parser(
        "upgrade",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[common_arguments],
        help="Apply upgrades to the existing PGQueuer database schema.",
    ).add_argument(
        "--dry-run",
        action="store_true",
        help=(
            "Print the SQL statements that would be executed without "
            "actually applying any changes to the database."
        ),
    )

    dashboardparser = subparsers.add_parser(
        "dashboard",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[common_arguments],
        help="Display a live dashboard showing job statistics.",
    )
    dashboardparser.add_argument(
        "-i",
        "--interval",
        help="Set the refresh interval in seconds for updating the dashboard display.",
        default=None,
        type=lambda x: timedelta(seconds=float(x)),
    )
    dashboardparser.add_argument(
        "-n",
        "--tail",
        help="Specify the number of the most recent log entries to display on the dashboard.",
        type=int,
        default=25,
    )
    dashboardparser.add_argument(
        "--table-format",
        default="pretty",
        help=(
            "Specify the format of the table used to display statistics. "
            "Options are provided by the tabulate library."
        ),
        choices=tabulate_formats,
    )

    listen_parser = subparsers.add_parser(
        "listen",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[common_arguments],
        help="Listen to a PostgreSQL NOTIFY channel for debug purposes.",
    )
    listen_parser.add_argument(
        "--channel",
        help="Specify the PostgreSQL NOTIFY channel to listen on for debug purposes.",
        default=queries.DBSettings().channel,
    )

    wm_parser = subparsers.add_parser(
        "run",
        description="Run a QueueManager from a factory function.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[common_arguments],
        help="Start a QueueManager to manage job queues and process jobs.",
    )
    wm_parser.add_argument(
        "--dequeue-timeout",
        help=(
            "Specify the timeout (in seconds) to wait to dequeue jobs. "
            "This determines how long the QueueManager will wait for new jobs before retrying."
        ),
        type=lambda s: timedelta(seconds=float(s)),
        default=timedelta(seconds=30.0),
    )
    wm_parser.add_argument(
        "--batch-size",
        help=(
            "Specify the number of jobs to dequeue in each batch. "
            "A larger batch size can increase throughput but may require more memory."
        ),
        type=int,
        default=10,
    )
    wm_parser.add_argument(
        "--retry-timer",
        help=(
            "Specify the time to wait (in seconds) before retrying jobs "
            "that have been picked but not processed."
        ),
        type=lambda s: timedelta(seconds=float(s)),
    )
    wm_parser.add_argument(
        "qm_factory",
        help="Path to the QueueManager factory function, e.g., 'myapp.create_queue_manager'.",
    )

    return parser.parse_args()


async def querier(
    driver: Literal["psy", "apg", "apgpool"],
    conninfo: str,
) -> queries.Queries:
    match driver:
        case "apg":
            import asyncpg

            return queries.Queries(
                db.AsyncpgDriver(
                    await asyncpg.connect(dsn=conninfo),
                )
            )
        case "apgpool":
            import asyncpg

            pool = await asyncpg.create_pool(dsn=conninfo)
            assert pool is not None
            return queries.Queries(db.AsyncpgPoolDriver(pool))

        case "psy":
            import psycopg

            return queries.Queries(
                db.PsycopgDriver(
                    await psycopg.AsyncConnection.connect(
                        conninfo=conninfo,
                        autocommit=True,
                    )
                )
            )

    raise NotImplementedError(driver)


async def main() -> None:  # noqa: C901
    parsed = cliparser()

    if "PGQUEUER_PREFIX" not in os.environ and isinstance(prefix := parsed.prefix, str) and prefix:
        os.environ["PGQUEUER_PREFIX"] = prefix

    dsn = parsed.pg_dsn or db.dsn(
        database=parsed.pg_database,
        password=parsed.pg_password,
        port=parsed.pg_port,
        user=parsed.pg_user,
        host=parsed.pg_host,
    )

    match parsed.command:
        case "install":
            print(queries.QueryBuilder().create_install_query())
            if not parsed.dry_run:
                await (await querier(parsed.driver, dsn)).install()
        case "uninstall":
            print(queries.QueryBuilder().create_uninstall_query())
            if not parsed.dry_run:
                await (await querier(parsed.driver, dsn)).uninstall()
        case "upgrade":
            print(f"\n{'-'*50}\n".join(queries.QueryBuilder().create_upgrade_queries()))
            if not parsed.dry_run:
                await (await querier(parsed.driver, dsn)).upgrade()
        case "dashboard":
            await fetch_and_display(
                await querier(parsed.driver, dsn),
                parsed.interval,
                parsed.tail,
                parsed.table_format,
            )
        case "listen":
            await display_pg_channel(
                (await querier(parsed.driver, dsn)).driver,
                models.PGChannel(parsed.channel),
            )
        case "run":
            await supervisor.runit(
                parsed.qm_factory,
                dequeue_timeout=parsed.dequeue_timeout,
                batch_size=parsed.batch_size,
                retry_timer=parsed.retry_timer,
            )



from __future__ import annotations

import argparse
import asyncio
import json
import os
import random
import signal
import sys
from contextlib import suppress
from datetime import datetime, timedelta, timezone
from itertools import count, groupby
from typing import Literal

from pydantic import AwareDatetime, BaseModel
from tqdm.asyncio import tqdm

from pgqueuer.cli import querier
from pgqueuer.db import dsn
from pgqueuer.models import Job
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


class BenchmarkResult(BaseModel):
    created_at: AwareDatetime
    driver: Literal["apg", "apgpool", "psy"]
    elapsed: timedelta
    github_ref_name: str
    rate: float
    steps: int


async def consumer(
    qm: QueueManager,
    batch_size: int,
    entrypoint_rps: list[float],
    concurrency_limits: list[int],
    bar: tqdm,
) -> None:
    assert len(entrypoint_rps) == 2
    async_rps, sync_rps = entrypoint_rps
    async_cl, sync_cl = concurrency_limits

    @qm.entrypoint(
        "asyncfetch",
        requests_per_second=async_rps,
        concurrency_limit=async_cl,
    )
    async def asyncfetch(job: Job) -> None:
        bar.update()

    @qm.entrypoint(
        "syncfetch",
        requests_per_second=sync_rps,
        concurrency_limit=sync_cl,
    )
    def syncfetch(job: Job) -> None:
        bar.update()

    await qm.run(batch_size=batch_size)


async def producer(
    shutdown: asyncio.Event,
    queries: Queries,
    batch_size: int,
    cnt: count,
) -> None:
    assert batch_size > 0
    entrypoints = ["syncfetch", "asyncfetch"] * batch_size
    while not shutdown.is_set():
        await queries.enqueue(
            random.sample(entrypoints, k=batch_size),
            [f"{next(cnt)}".encode() for _ in range(batch_size)],
            [0] * batch_size,
        )


def cli_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="PGQueuer benchmark tool.")

    parser.add_argument(
        "-d",
        "--driver",
        default="apg",
        help="Postgres driver to be used asyncpg (apg) or psycopg (psy).",
        choices=["apg", "apgpool", "psy"],
    )

    parser.add_argument(
        "-t",
        "--timer",
        type=lambda x: timedelta(seconds=float(x)),
        default=timedelta(seconds=30),
        help="Run the benchmark for a specified number of seconds. Default is 10.",
    )

    parser.add_argument(
        "-dq",
        "--dequeue",
        type=int,
        default=5,
        help="Number of concurrent dequeue tasks. Default is 5.",
    )
    parser.add_argument(
        "-dqbs",
        "--dequeue-batch-size",
        type=int,
        default=10,
        help="Batch size for dequeue tasks. Default is 10.",
    )

    parser.add_argument(
        "-eq",
        "--enqueue",
        type=int,
        default=1,
        help="Number of concurrent enqueue tasks. Default is 1.",
    )
    parser.add_argument(
        "-eqbs",
        "--enqueue-batch-size",
        type=int,
        default=10,
        help="Batch size for enqueue tasks. Default is 10.",
    )
    parser.add_argument(
        "-rps",
        "--requests-per-second",
        nargs="+",
        default=[float("inf"), float("inf")],
        help="RPS for endporints given as a list, defautl is 'inf'.",
    )
    parser.add_argument(
        "-ci",
        "--concurrency-limit",
        nargs="+",
        default=[sys.maxsize, sys.maxsize],
        help=f"Concurrency limit for endporints given as a list, defautl is '{sys.maxsize}'.",
    )
    parser.add_argument(
        "-o",
        "--output-json",
        type=str,
        default=None,
        help="Path to the output JSON file for benchmark metrics.",
    )
    return parser.parse_args()


async def main(args: argparse.Namespace) -> None:
    print(f"""Settings:
Timer:                  {args.timer.total_seconds()} seconds
Dequeue:                {args.dequeue}
Dequeue Batch Size:     {args.dequeue_batch_size}
Enqueue:                {args.enqueue}
Enqueue Batch Size:     {args.enqueue_batch_size}
""")

    await (await querier(args.driver, dsn())).clear_log()
    await (await querier(args.driver, dsn())).clear_queue()

    shutdown = asyncio.Event()
    qms = list[QueueManager]()
    tqdm_format_dict = {}

    async def enqueue(shutdown: asyncio.Event) -> None:
        cnt = count()
        producers = [
            producer(
                shutdown,
                await querier(args.driver, dsn()),
                int(args.enqueue_batch_size),
                cnt,
            )
            for _ in range(args.enqueue)
        ]
        await asyncio.gather(*producers)

    async def dequeue(qms: list[QueueManager]) -> None:
        queries = [await querier(args.driver, dsn()) for _ in range(args.dequeue)]
        for q in queries:
            qms.append(QueueManager(q.driver))

        with tqdm(
            ascii=True,
            unit=" job",
            unit_scale=True,
            file=sys.stdout,
        ) as bar:
            consumers = [
                consumer(
                    qm=q,
                    batch_size=int(args.dequeue_batch_size),
                    entrypoint_rps=[float(x) for x in args.requests_per_second],
                    concurrency_limits=[int(x) for x in args.concurrency_limit],
                    bar=bar,
                )
                for q in qms
            ]
            await asyncio.gather(*consumers)
            tqdm_format_dict.update(bar.format_dict)

    async def dequeue_shutdown_timer(
        qms: list[QueueManager],
        shutdown: asyncio.Event,
    ) -> None:
        _, pending = await asyncio.wait(
            (
                asyncio.create_task(asyncio.sleep(args.timer.total_seconds())),
                asyncio.create_task(shutdown.wait()),
            ),
            return_when=asyncio.FIRST_COMPLETED,
        )
        # Stop producers
        shutdown.set()
        # Stop consumers
        for q in qms:
            q.shutdown.set()

        for p in pending:
            p.cancel()

    def graceful_shutdown(signum: int, frame: object) -> None:
        shutdown.set()
        for qm in qms:
            qm.shutdown.set()

    signal.signal(signal.SIGINT, graceful_shutdown)  # Handle Ctrl-C
    signal.signal(signal.SIGTERM, graceful_shutdown)  # Handle termination request

    await asyncio.gather(
        dequeue(qms),
        enqueue(shutdown),
        dequeue_shutdown_timer(qms, shutdown),
    )

    qsize = await (await querier(args.driver, dsn())).queue_size()
    print("Queue size:")
    for status, items in groupby(sorted(qsize, key=lambda x: x.status), key=lambda x: x.status):
        print(f"  {status} {sum(x.count for x in items)}")
    if not qsize:
        print("  0")

    if tqdm_format_dict and args.output_json:
        with open(args.output_json, "w") as f:
            json.dump(
                BenchmarkResult(
                    driver=args.driver,
                    created_at=datetime.now(timezone.utc),
                    elapsed=tqdm_format_dict["elapsed"],
                    rate=tqdm_format_dict["rate"],
                    steps=tqdm_format_dict["n"],
                    github_ref_name=os.environ.get("REF_NAME", ""),
                ).model_dump(mode="json"),
                f,
            )


if __name__ == "__main__":
    with suppress(KeyboardInterrupt):
        asyncio.run(main(cli_parser()))



from __future__ import annotations

import asyncio
import uuid
from collections import defaultdict
from datetime import timedelta
from itertools import chain

import pytest

from pgqueuer.db import Driver
from pgqueuer.models import Job
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


async def raises(lock: asyncio.Lock) -> None:
    if lock.locked():
        raise ValueError("Locked")


async def consumer(
    qm: QueueManager,
    lock: asyncio.Lock,
) -> defaultdict[str, list[Job]]:
    consumed = defaultdict[str, list[Job]](list)

    @qm.entrypoint("serialized_dispatch_true", serialized_dispatch=True)
    async def serialized_dispatch_true(job: Job) -> None:
        consumed["serialized_dispatch_true"].append(job)
        await asyncio.sleep(0.1)
        await raises(lock)

    @qm.entrypoint("serialized_dispatch_false", serialized_dispatch=False)
    async def fetch(job: Job) -> None:
        consumed["serialized_dispatch_false"].append(job)
        await asyncio.sleep(0.1)

    await qm.run(dequeue_timeout=timedelta(seconds=0))
    return consumed


async def enqueue(
    queries: Queries,
    size: int,
) -> None:
    assert size > 0
    await queries.enqueue(
        ["serialized_dispatch_true"] * size,
        [f"{n}".encode() for n in range(size)],
        [0] * size,
    )
    await queries.enqueue(
        ["serialized_dispatch_false"] * size,
        [f"{n}".encode() for n in range(size)],
        [0] * size,
    )


async def locked_consumer(
    qm: QueueManager,
) -> defaultdict[str, list[Job]]:
    consumed = defaultdict[str, list[Job]](list)

    @qm.entrypoint("serialized_dispatch_true", serialized_dispatch=True)
    async def serialized_dispatch_true(job: Job) -> None:
        consumed["serialized_dispatch_true"].append(job)
        await asyncio.sleep(0.1)
        raise ValueError("Locked")  # Always raise to simulate locked state

    await qm.run(dequeue_timeout=timedelta(seconds=0))
    return consumed


@pytest.mark.parametrize("n_consumers", (1, 2, 4, 32))
async def test_serialized_dispatch(
    n_consumers: int,
    apgdriver: Driver,
    n_tasks: int = 500,
    wait: int = 1,
) -> None:
    lock = asyncio.Lock()
    await enqueue(Queries(apgdriver), size=n_tasks)
    qms = [QueueManager(apgdriver) for _ in range(n_consumers)]

    async def dequeue() -> None:
        consumers = [consumer(q, lock) for q in qms]
        await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    await asyncio.gather(timer(), dequeue())


@pytest.mark.parametrize("n_consumers", (2, 4))
async def test_serialized_dispatch_mixed(
    n_consumers: int,
    apgdriver: Driver,
    n_tasks: int = 10_000,
    wait: int = 1,  # Set wait time to 1 second
) -> None:
    lock = asyncio.Lock()
    await enqueue(Queries(apgdriver), size=n_tasks)
    qms = [QueueManager(apgdriver) for _ in range(n_consumers)]

    async def dequeue() -> list[defaultdict[str, list[Job]]]:
        consumers = [consumer(q, lock) for q in qms]
        return await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    _, consumed = await asyncio.gather(timer(), dequeue())
    serialized_dispatch_true = list(
        chain.from_iterable(x["serialized_dispatch_true"] for x in consumed)
    )
    serialized_dispatch_false = list(
        chain.from_iterable(x["serialized_dispatch_false"] for x in consumed)
    )

    # Add debug information to monitor job counts
    assert len(serialized_dispatch_true) > 0, "No jobs processed for serialized_dispatch_true"
    assert len(serialized_dispatch_false) > 0, "No jobs processed for serialized_dispatch_false"


@pytest.mark.parametrize("n_consumers", (1, 2, 4))
async def test_no_jobs_processed_when_locked(
    n_consumers: int,
    apgdriver: Driver,
    n_tasks: int = 100,
    wait: int = 3,  # Added wait time for timed shutdown
) -> None:
    # Set one job to picked state to simulate an ongoing process
    queries = Queries(apgdriver)
    await enqueue(queries, size=n_tasks)
    picked_job = await queries.dequeue(
        1,
        {"serialized_dispatch_true": (timedelta(seconds=30), True, 0)},
        queue_manager_id=uuid.uuid4(),
    )
    assert len(picked_job) == 1, "Failed to pick a job for locking"

    qms = [QueueManager(apgdriver) for _ in range(n_consumers)]

    async def dequeue() -> list[defaultdict[str, list[Job]]]:
        consumers = [locked_consumer(q) for q in qms]
        return await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    _, consumed = await asyncio.gather(timer(), dequeue())
    serialized_dispatch_true = list(
        chain.from_iterable(x["serialized_dispatch_true"] for x in consumed)
    )
    assert (
        len(serialized_dispatch_true) == 0
    ), "Jobs should not have been processed while lock is acquired"


@pytest.mark.parametrize("n_consumers", (1, 2, 4))
async def test_mixed_serialized_and_concurrent_processing(
    n_consumers: int,
    apgdriver: Driver,
    n_tasks: int = 1000,
    wait: int = 1,  # Set wait time to 1 second
) -> None:
    lock = asyncio.Lock()
    await enqueue(Queries(apgdriver), size=n_tasks)
    qms = [QueueManager(apgdriver) for _ in range(n_consumers)]

    async def dequeue() -> list[defaultdict[str, list[Job]]]:
        consumers = [consumer(q, lock) for q in qms]
        return await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    _, consumed = await asyncio.gather(timer(), dequeue())
    serialized_dispatch_true = list(
        chain.from_iterable(x["serialized_dispatch_true"] for x in consumed)
    )
    serialized_dispatch_false = list(
        chain.from_iterable(x["serialized_dispatch_false"] for x in consumed)
    )
    # Ensure some jobs were processed with serialization and some concurrently
    assert len(serialized_dispatch_true) > 0
    assert len(serialized_dispatch_false) > 0


@pytest.mark.parametrize("n_consumers", (1,))
async def test_single_consumer_serialized_behavior(
    n_consumers: int,
    apgdriver: Driver,
    n_tasks: int = 50,
    wait: int = 1,  # Set wait time to 1 second
) -> None:
    lock = asyncio.Lock()
    await enqueue(Queries(apgdriver), size=n_tasks)
    qms = [QueueManager(apgdriver) for _ in range(n_consumers)]

    async def dequeue() -> list[defaultdict[str, list[Job]]]:
        consumers = [consumer(q, lock) for q in qms]
        return await asyncio.gather(*consumers)

    async def timer() -> None:
        await asyncio.sleep(wait)
        for q in qms:
            q.shutdown.set()

    _, consumed = await asyncio.gather(timer(), dequeue())
    serialized_dispatch_true = list(
        chain.from_iterable(x["serialized_dispatch_true"] for x in consumed)
    )
    # Verify that all jobs were processed sequentially by a single consumer
    assert len(serialized_dispatch_true) == n_tasks



from __future__ import annotations

import importlib
import signal
from datetime import timedelta
from typing import Awaitable, Callable, TypeAlias

from . import qm

QM_FACTORY: TypeAlias = Callable[[], Awaitable[qm.QueueManager]]


def load_queue_manager_factory(factory_path: str) -> QM_FACTORY:
    """
    Load the QueueManager factory function from a given module path.

    Dynamically imports the specified module and retrieves the factory function
    used to create a QueueManager instance. The factory function should be an
    asynchronous callable that returns a QueueManager.

    Args:
        factory_path (str): The full module path to the factory function,
            e.g., 'myapp.create_queue_manager'.

    Returns:
        QM_FACTORY: A callable that returns an awaitable QueueManager instance.
    """
    module_name, factory_name = factory_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, factory_name)


async def runit(
    factory_fn: str,
    dequeue_timeout: timedelta,
    batch_size: int,
    retry_timer: timedelta | None,
) -> None:
    """
    Instantiate and run a QueueManager using the provided factory function.

    This function handles the setup and lifecycle management of the QueueManager,
    including signal handling for graceful shutdown. It loads the QueueManager
    factory, creates an instance, sets up signal handlers, and starts the
    QueueManager's run loop with the specified parameters.

    Args:
        factory_fn (str): The module path to the QueueManager factory function,
            e.g., 'myapp.create_queue_manager'.
        dequeue_timeout (timedelta): The timeout duration for dequeuing jobs.
        batch_size (int): The number of jobs to retrieve in each batch.
        retry_timer (timedelta | None): The duration after which to retry 'picked' jobs
            that may have stalled. If None, retry logic is disabled.
    """
    qm_factory = load_queue_manager_factory(factory_fn)
    qm = await qm_factory()  # Create the QueueManager instance

    def graceful_shutdown(signum: int, frame: object) -> None:
        """
        Handle incoming signals to perform a graceful shutdown.

        When a termination signal is received (e.g., SIGINT or SIGTERM), this function
        sets the 'shutdown' event in the QueueManager to initiate a controlled shutdown
        process, allowing tasks to complete cleanly.

        Args:
            signum (int): The signal number received.
            frame (object): The current stack frame (unused).
        """
        print(f"Received signal {signum}, shutting down...")
        qm.shutdown.set()

    # Setup signal handlers for graceful shutdown of the QueueManager.
    signal.signal(signal.SIGINT, graceful_shutdown)  # Handle Ctrl-C
    signal.signal(signal.SIGTERM, graceful_shutdown)  # Handle termination request

    # Start the QueueManager's operation
    await qm.run(
        dequeue_timeout=dequeue_timeout,
        batch_size=batch_size,
        retry_timer=retry_timer,
    )



import asyncio
from typing import AsyncGenerator

import asyncpg
import pytest

from pgqueuer.db import AsyncpgDriver, Driver, dsn
from pgqueuer.queries import Queries


@pytest.fixture(scope="function")
async def apgdriver() -> AsyncGenerator[AsyncpgDriver, None]:
    conn = await asyncpg.connect(dsn=dsn())
    try:
        yield AsyncpgDriver(conn)
    finally:
        await conn.close()


@pytest.fixture(scope="function", autouse=True)
async def trucate_tables(apgdriver: Driver) -> None:
    await asyncio.gather(
        Queries(apgdriver).clear_log(),
        Queries(apgdriver).clear_queue(),
    )



from typing import AsyncGenerator, Generator

import asyncpg
import pytest
from fastapi import Depends, FastAPI
from fastapi.testclient import TestClient

from pgqueuer.db import AsyncpgDriver, Driver
from pgqueuer.queries import Queries


async def get_driver() -> AsyncGenerator[Driver, None]:
    conn = await asyncpg.connect()
    try:
        yield AsyncpgDriver(conn)
    finally:
        await conn.close()


@pytest.fixture
def client() -> Generator[TestClient, None, None]:
    app = FastAPI()

    @app.post("/enqueue")
    async def enqueue_job(
        entrypoint: str,
        payload: str,
        priority: int = 0,
        driver: AsyncpgDriver = Depends(get_driver),
    ) -> dict:
        queries = Queries(driver)
        ids = await queries.enqueue(entrypoint, payload.encode(), priority)
        return {"job-ids": ids}

    @app.get("/queue-size")
    async def get_queue_size(
        driver: AsyncpgDriver = Depends(get_driver),
    ) -> dict:
        queries = Queries(driver)
        size = await queries.queue_size()
        return {"queue_size": size}

    with TestClient(app) as client:
        yield client


def test_enqueue_job(client: TestClient) -> None:
    response = client.post(
        "/enqueue",
        params={
            "entrypoint": "test_function",
            "payload": "data",
            "priority": 5,
        },
    )
    assert response.status_code == 200
    assert isinstance(response.json()["job-ids"], list)


def test_queue_size(client: TestClient) -> None:
    response = client.get("/queue-size")
    assert response.status_code == 200
    assert "queue_size" in response.json()



from datetime import datetime, timedelta

from pgqueuer.helpers import perf_counter_dt, retry_timer_buffer_timeout, timeout_with_jitter


async def test_perf_counter_dt() -> None:
    assert isinstance(perf_counter_dt(), datetime)
    assert perf_counter_dt().tzinfo is not None


def test_heartbeat_buffer_timeout_empty_list() -> None:
    dts = list[timedelta]()
    expected = timedelta(hours=24)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_all_dts_less_than_or_equal_to_t0() -> None:
    dts = [timedelta(seconds=-1), timedelta(seconds=0)]
    expected = timedelta(hours=24)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_positive_dts() -> None:
    dts = [timedelta(seconds=10), timedelta(seconds=5)]
    expected = timedelta(seconds=5)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_mixed_dts() -> None:
    dts = [timedelta(seconds=-5), timedelta(seconds=10)]
    expected = timedelta(seconds=10)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_custom_t0() -> None:
    dts = [timedelta(seconds=4), timedelta(seconds=6)]
    expected = timedelta(seconds=6)
    assert retry_timer_buffer_timeout(dts, _t0=timedelta(seconds=5)) == expected


def test_heartbeat_buffer_timeout_custom_default() -> None:
    dts = list[timedelta]()
    expected = timedelta(hours=48)
    assert retry_timer_buffer_timeout(dts, _default=timedelta(hours=48)) == expected


def test_delay_within_jitter_range() -> None:
    base_timeout = timedelta(seconds=10)
    delay_multiplier = 2.0
    jitter_span = (0.8, 1.2)

    # Call the function multiple times to check the jitter range
    for _ in range(100):
        delay = timeout_with_jitter(base_timeout, delay_multiplier, jitter_span)
        base_delay = base_timeout.total_seconds() * delay_multiplier
        assert base_delay * jitter_span[0] <= delay.total_seconds() <= base_delay * jitter_span[1]


def test_delay_is_timedelta() -> None:
    base_timeout = timedelta(seconds=5)
    delay_multiplier = 1.5
    delay = timeout_with_jitter(base_timeout, delay_multiplier)
    assert isinstance(delay, timedelta)


def test_custom_jitter_range() -> None:
    base_timeout = timedelta(seconds=8)
    delay_multiplier = 1.0
    jitter_span = (0.5, 1.5)

    # Call the function multiple times to check the custom jitter range
    for _ in range(100):
        delay = timeout_with_jitter(base_timeout, delay_multiplier, jitter_span)
        base_delay = base_timeout.total_seconds() * delay_multiplier
        assert base_delay * jitter_span[0] <= delay.total_seconds() <= base_delay * jitter_span[1]
